# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from collections import namedtuple
from time import clock

from marionette_driver import By, expected, Wait
from marionette_harness import Marionette

from video_puppeteer import VideoPuppeteer, TimeRanges
from external_media_tests.utils import verbose_until


class TwitchPuppeteer(VideoPuppeteer):
    """
    Wrapper around a Twitch .player element.

    Note that the twitch stream needs to be playing for the puppeteer to work
    correctly. Since twitch will load a player element for streams that are
    not currently playing, the puppeteer will still detect a player element,
    however the time ranges will not increase, and tests will fail.

    Compared to video puppeteer, this class has the benefit of accessing the
    twitch player element as well as the video element. The twitch player
    element reports additional information to aid in testing -- such as if an
    ad is playing. However, the attributes of this element do not appear to be
    documented, which may leave them vulnerable to undocumented changes in
    behaviour.
    """
    _player_var_script = (
        'var player_data_screen = '
        'arguments[1].wrappedJSObject.getAttribute("data-screen");'
    )
    """
    A string containing JS that will assign player state to
    variables. This is similar to `_video_var_script` from
    `VideoPuppeteer`. See `_video_var_script` for more information on the
    motivation for this method.
    """

    def __init__(self, marionette, url, autostart=True,
                 set_duration=10.0, **kwargs):
        self.player = None
        self._last_seen_player_state = None
        super(TwitchPuppeteer,
              self).__init__(marionette, url, set_duration=set_duration,
                             autostart=False, **kwargs)
        wait = Wait(self.marionette, timeout=30)
        with self.marionette.using_context(Marionette.CONTEXT_CONTENT):
            verbose_until(wait, self,
                          expected.element_present(By.CLASS_NAME,
                                                   'player'))
            self.player = self.marionette.find_element(By.CLASS_NAME,
                                                       'player')
            self.marionette.execute_script("log('.player "
                                           "element obtained');")
            if autostart:
                self.start()

    def _update_expected_duration(self):
        if 0 < self._set_duration:
            self.expected_duration = self._set_duration
        else:
            # If we don't have a set duration we don't know how long is left
            # in the stream.
            self.expected_duration = float("inf")

    def _calculate_remaining_time(self, played_ranges):
        """
        Override of video_puppeteer's _calculate_remaining_time. See that
        method for primary documentation.

        This override is in place to adjust how remaining time is handled.
        Twitch ads can cause small stutters which result in multiple played
        ranges, despite no seeks being initiated by the tests. As such, when
        calculating the remaining time, the start time is the min of all
        played start times, and the end time is the max of played end times.
        This being sensible behaviour relies on the tests not attempting seeks.

        :param played_ranges: A TimeRanges object containing played ranges.
        :return: The remaining time expected for this puppeteer.
        """
        min_played_time = min(
            [played_ranges.start(i) for i in range(0, played_ranges.length)])
        max_played_time = max(
            [played_ranges.end(i) for i in range(0, played_ranges.length)])
        played_duration = max_played_time - min_played_time
        return self.expected_duration - played_duration

    def _execute_twitch_script(self, script):
        """
        Execute JS script in content context with access to video element and
        Twitch .player element.

        :param script: script to be executed.

        :return: value returned by script
        """
        with self.marionette.using_context(Marionette.CONTEXT_CONTENT):
            return self.marionette.execute_script(script,
                                                  script_args=[self.video,
                                                               self.player])

    @staticmethod
    def _twitch_state_named_tuple():
        """
        Create a named tuple class that can be used to store state snapshots
        of the wrapped twitch player. The fields in the tuple should be used
        as follows:

        player_data_screen: the current displayed content, appears to be set
        to nothing if no ad has been played, 'advertisement' during ad
        playback, and 'content' following ad playback.
        """
        return namedtuple('player_state_info',
                          ['player_data_screen'])

    def _create_player_state_info(self, **player_state_info_kwargs):
        """
        Create an instance of the state info named tuple. This function
        expects a dictionary containing the following keys:
        player_data_screen.

        For more information on the above keys and their values see
        `_twitch_state_named_tuple`.

        :return: A named tuple 'player_state_info', derived from arguments and
        state information from the puppeteer.
        """
        # Create player snapshot
        state_info = self._twitch_state_named_tuple()
        return state_info(**player_state_info_kwargs)

    @property
    def _fetch_state_script(self):
        if not self._fetch_state_script_string:
            self._fetch_state_script_string = (
                self._video_var_script +
                self._player_var_script +
                'return ['
                'baseURI,'
                'currentTime,'
                'duration,'
                '[buffered.length, bufferedRanges],'
                '[played.length, playedRanges],'
                'totalFrames,'
                'droppedFrames,'
                'corruptedFrames,'
                'player_data_screen];')
        return self._fetch_state_script_string

    def _refresh_state(self):
        """
        Refresh the snapshot of the underlying video and player state. We do
        this all in one so that the state doesn't change in between queries.

        We also store information thouat can be derived from the snapshotted
        information, such as lag. This is stored in the last seen state to
        stress that it's based on the snapshot.
        """
        values = self._execute_twitch_script(self._fetch_state_script)
        video_keys = ['base_uri', 'current_time', 'duration',
                      'raw_buffered_ranges', 'raw_played_ranges',
                      'total_frames', 'dropped_frames', 'corrupted_frames']
        player_keys = ['player_data_screen']
        # Get video state
        self._last_seen_video_state = (
            self._create_video_state_info(**dict(
                zip(video_keys, values[:len(video_keys)]))))
        # Get player state
        self._last_seen_player_state = (
            self._create_player_state_info(**dict(
                zip(player_keys, values[-len(player_keys):]))))

    def __str__(self):
        messages = [super(TwitchPuppeteer, self).__str__()]
        if not self.player:
            messages += ['\t.player: None']
            return '\n'.join(messages)
        if not self._last_seen_player_state:
            messages += ['\t.player: No last seen state']
            return '\n'.join(messages)
        messages += ['.player: {']
        for field in self._last_seen_player_state._fields:
            # For compatibility with different test environments we force
            # ascii
            field_ascii = (
                unicode(getattr(self._last_seen_player_state, field)).encode(
                    'ascii', 'replace'))
            messages += [('\t{}: {}'.format(field, field_ascii))]
        messages += '}'
        return '\n'.join(messages)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from external_media_harness.testcase import (
    MediaTestCase,
    VideoPlaybackTestsMixin
)


class TestVideoPlayback(MediaTestCase, VideoPlaybackTestsMixin):

    # Tests are actually implemented in VideoPlaybackTestsMixin.

    pass



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import re

from collections import namedtuple
from json import loads
from time import sleep

from marionette_driver import By, expected, Wait
from marionette_driver.errors import TimeoutException, NoSuchElementException
from marionette_harness import Marionette

from video_puppeteer import VideoPuppeteer, VideoException
from external_media_tests.utils import verbose_until


class YouTubePuppeteer(VideoPuppeteer):
    """
    Wrapper around a YouTube .html5-video-player element.

    Can be used with youtube videos or youtube videos at embedded URLS. E.g.
    both https://www.youtube.com/watch?v=AbAACm1IQE0 and
    https://www.youtube.com/embed/AbAACm1IQE0 should work.

    Using an embedded video has the advantage of not auto-playing more videos
    while a test is running.

    Compared to video puppeteer, this class has the benefit of accessing the
    youtube player object as well as the video element. The YT player will
    report information for the underlying video even if an add is playing (as
    opposed to the video element behaviour, which will report on whatever
    is play at the time of query), and can also report if an ad is playing.

    Partial reference: https://developers.google.com/youtube/iframe_api_reference.
    This reference is useful for site-specific features such as interacting
    with ads, or accessing YouTube's debug data.
    """

    _player_var_script = (
        'var player_duration = arguments[1].wrappedJSObject.getDuration();'
        'var player_current_time = '
        'arguments[1].wrappedJSObject.getCurrentTime();'
        'var player_playback_quality = '
        'arguments[1].wrappedJSObject.getPlaybackQuality();'
        'var player_movie_id = '
        'arguments[1].wrappedJSObject.getVideoData()["video_id"];'
        'var player_movie_title = '
        'arguments[1].wrappedJSObject.getVideoData()["title"];'
        'var player_url = '
        'arguments[1].wrappedJSObject.getVideoUrl();'
        'var player_state = '
        'arguments[1].wrappedJSObject.getPlayerState();'
        'var player_ad_state = arguments[1].wrappedJSObject.getAdState();'
        'var player_breaks_count = '
        'arguments[1].wrappedJSObject.getOption("ad", "breakscount");'
    )
    """
    A string containing JS that will assign player state to
    variables. This is similar to `_video_var_script` from
    `VideoPuppeteer`. See `_video_var_script` for more information on the
    motivation for this method.

    This script assigns a subset of the vars used later by the
    `_yt_state_named_tuple` function. Please see that functions
    documentation for further information on these variables.
    """

    _yt_player_state = {
        'UNSTARTED': -1,
        'ENDED': 0,
        'PLAYING': 1,
        'PAUSED': 2,
        'BUFFERING': 3,
        'CUED': 5
    }
    _yt_player_state_name = {v: k for k, v in _yt_player_state.items()}
    _time_pattern = re.compile('(?P<minute>\d+):(?P<second>\d+)')

    def __init__(self, marionette, url, autostart=True, **kwargs):
        self.player = None
        self._last_seen_player_state = None
        super(YouTubePuppeteer,
              self).__init__(marionette, url,
                             video_selector='.html5-video-player video',
                             autostart=False,
                             **kwargs)
        wait = Wait(self.marionette, timeout=30)
        with self.marionette.using_context(Marionette.CONTEXT_CONTENT):
            verbose_until(wait, self,
                          expected.element_present(By.CLASS_NAME,
                                                   'html5-video-player'))
            self.player = self.marionette.find_element(By.CLASS_NAME,
                                                       'html5-video-player')
            self.marionette.execute_script("log('.html5-video-player "
                                           "element obtained');")
        # When an ad is playing, self.player_duration indicates the duration
        # of the spliced-in ad stream, not the duration of the main video, so
        # we attempt to skip the ad first.
        for attempt in range(5):
            sleep(1)
            self.process_ad()
            if (self._last_seen_player_state.player_ad_inactive and
                    self._last_seen_video_state.duration and not
                    self._last_seen_player_state.player_buffering):
                break
        self._update_expected_duration()
        if autostart:
            self.start()

    def player_play(self):
        """
        Play via YouTube API.
        """
        self._execute_yt_script('arguments[1].wrappedJSObject.playVideo();')

    def player_pause(self):
        """
        Pause via YouTube API.
        """
        self._execute_yt_script('arguments[1].wrappedJSObject.pauseVideo();')

    def _player_measure_progress(self):
        """
        Determine player progress. Refreshes state.

        :return: Playback progress in seconds via YouTube API with snapshots.
        """
        self._refresh_state()
        initial = self._last_seen_player_state.player_current_time
        sleep(1)
        self._refresh_state()
        return self._last_seen_player_state.player_current_time - initial

    def _get_player_debug_dict(self):
        text = self._execute_yt_script('return arguments[1].'
                                       'wrappedJSObject.getDebugText();')
        if text:
            try:
                return loads(text)
            except ValueError:
                self.marionette.log('Error loading json: DebugText',
                                    level='DEBUG')

    def _execute_yt_script(self, script):
        """
        Execute JS script in content context with access to video element and
        YouTube .html5-video-player element.

        :param script: script to be executed.

        :return: value returned by script
        """
        with self.marionette.using_context(Marionette.CONTEXT_CONTENT):
            return self.marionette.execute_script(script,
                                                  script_args=[self.video,
                                                               self.player])

    def _check_if_ad_ended(self):
        self._refresh_state()
        return self._last_seen_player_state.player_ad_ended

    def process_ad(self):
        """
        Wait for this ad to finish. Refreshes state.
        """
        self._refresh_state()
        if self._last_seen_player_state.player_ad_inactive:
            return
        ad_timeout = (self._search_ad_duration() or 30) + 5
        wait = Wait(self, timeout=ad_timeout, interval=1)
        try:
            self.marionette.log('process_ad: waiting {} s for ad'
                                .format(ad_timeout))
            verbose_until(wait,
                          self,
                          YouTubePuppeteer._check_if_ad_ended,
                          "Check if ad ended")
        except TimeoutException:
            self.marionette.log('Waiting for ad to end timed out',
                                level='WARNING')

    def _search_ad_duration(self):
        """
        Try and determine ad duration. Refreshes state.

        :return: ad duration in seconds, if currently displayed in player
        """
        self._refresh_state()
        if not (self._last_seen_player_state.player_ad_playing or
                self._player_measure_progress() == 0):
            return None
        if (self._last_seen_player_state.player_ad_playing and
                self._last_seen_video_state.duration):
            return self._last_seen_video_state.duration
        selector = '.html5-video-player .videoAdUiAttribution'
        wait = Wait(self.marionette, timeout=5)
        try:
            with self.marionette.using_context(Marionette.CONTEXT_CONTENT):
                wait.until(expected.element_present(By.CSS_SELECTOR,
                                                    selector))
                countdown = self.marionette.find_element(By.CSS_SELECTOR,
                                                         selector)
                ad_time = self._time_pattern.search(countdown.text)
                if ad_time:
                    ad_minutes = int(ad_time.group('minute'))
                    ad_seconds = int(ad_time.group('second'))
                    return 60 * ad_minutes + ad_seconds
        except (TimeoutException, NoSuchElementException):
            self.marionette.log('Could not obtain '
                                'element: {}'.format(selector),
                                level='WARNING')
        return None

    def _player_stalled(self):
        """
        Checks if the player has stalled. Refreshes state.

        :return: True if playback is not making progress for 4-9 seconds. This
         excludes ad breaks. Note that the player might just be busy with
         buffering due to a slow network.
        """

        # `current_time` stands still while ad is playing
        def condition():
            # no ad is playing and current_time stands still
            return (not self._last_seen_player_state.player_ad_playing and
                    self._measure_progress() < 0.1 and
                    self._player_measure_progress() < 0.1 and
                    (self._last_seen_player_state.player_playing or
                     self._last_seen_player_state.player_buffering))

        if condition():
            sleep(2)
            self._refresh_state()
            if self._last_seen_player_state.player_buffering:
                sleep(5)
                self._refresh_state()
            return condition()
        else:
            return False

    @staticmethod
    def _yt_state_named_tuple():
        """
        Create a named tuple class that can be used to store state snapshots
        of the wrapped youtube player. The fields in the tuple should be used
        as follows:

        player_duration: the duration as fetched from the wrapped player.
        player_current_time: the current playback time as fetched from the
        wrapped player.
        player_remaining_time: the remaining time as calculated based on the
        puppeteers expected time and the players current time.
        player_playback_quality: the playback quality as fetched from the
        wrapped player. See:
        https://developers.google.com/youtube/js_api_reference#Playback_quality
        player_movie_id: the movie id fetched from the wrapped player.
        player_movie_title: the title fetched from the wrapped player.
        player_url: the self reported url fetched from the wrapped player.
        player_state: the current state of playback as fetch from the wrapped
        player. See:
        https://developers.google.com/youtube/js_api_reference#Playback_status
        player_unstarted, player_ended, player_playing, player_paused,
        player_buffering, and player_cued: these are all shortcuts to the
        player state, only one should be true at any time.
        player_ad_state: as player_state, but reports for the current ad.
        player_ad_state, player_ad_inactive, player_ad_playing, and
        player_ad_ended: these are all shortcuts to the ad state, only one
        should be true at any time.
        player_breaks_count: the number of ads as fetched from the wrapped
        player. This includes both played and unplayed ads, and includes
        streaming ads as well as pop up ads.

        :return: A 'player_state_info' named tuple class.
        """
        return namedtuple('player_state_info',
                          ['player_duration',
                           'player_current_time',
                           'player_remaining_time',
                           'player_playback_quality',
                           'player_movie_id',
                           'player_movie_title',
                           'player_url',
                           'player_state',
                           'player_unstarted',
                           'player_ended',
                           'player_playing',
                           'player_paused',
                           'player_buffering',
                           'player_cued',
                           'player_ad_state',
                           'player_ad_inactive',
                           'player_ad_playing',
                           'player_ad_ended',
                           'player_breaks_count'
                           ])

    def _create_player_state_info(self, **player_state_info_kwargs):
        """
        Create an instance of the state info named tuple. This function
        expects a dictionary containing the following keys:
        player_duration, player_current_time, player_playback_quality,
        player_movie_id, player_movie_title, player_url, player_state,
        player_ad_state, and player_breaks_count.

        For more information on the above keys and their values see
        `_yt_state_named_tuple`.

        :return: A named tuple 'player_state_info', derived from arguments and
        state information from the puppeteer.
        """
        player_state_info_kwargs['player_remaining_time'] = (
            self.expected_duration -
            player_state_info_kwargs['player_current_time'])
        # Calculate player state convenience info
        player_state = player_state_info_kwargs['player_state']
        player_state_info_kwargs['player_unstarted'] = (
            player_state == self._yt_player_state['UNSTARTED'])
        player_state_info_kwargs['player_ended'] = (
            player_state == self._yt_player_state['ENDED'])
        player_state_info_kwargs['player_playing'] = (
            player_state == self._yt_player_state['PLAYING'])
        player_state_info_kwargs['player_paused'] = (
            player_state == self._yt_player_state['PAUSED'])
        player_state_info_kwargs['player_buffering'] = (
            player_state == self._yt_player_state['BUFFERING'])
        player_state_info_kwargs['player_cued'] = (
            player_state == self._yt_player_state['CUED'])
        # Calculate ad state convenience info
        player_ad_state = player_state_info_kwargs['player_ad_state']
        player_state_info_kwargs['player_ad_inactive'] = (
            player_ad_state == self._yt_player_state['UNSTARTED'])
        player_state_info_kwargs['player_ad_playing'] = (
            player_ad_state == self._yt_player_state['PLAYING'])
        player_state_info_kwargs['player_ad_ended'] = (
            player_ad_state == self._yt_player_state['ENDED'])
        # Create player snapshot
        state_info = self._yt_state_named_tuple()
        return state_info(**player_state_info_kwargs)

    @property
    def _fetch_state_script(self):
        if not self._fetch_state_script_string:
            self._fetch_state_script_string = (
                self._video_var_script +
                self._player_var_script +
                'return ['
                'baseURI,'
                'currentTime,'
                'duration,'
                '[buffered.length, bufferedRanges],'
                '[played.length, playedRanges],'
                'totalFrames,'
                'droppedFrames,'
                'corruptedFrames,'
                'player_duration,'
                'player_current_time,'
                'player_playback_quality,'
                'player_movie_id,'
                'player_movie_title,'
                'player_url,'
                'player_state,'
                'player_ad_state,'
                'player_breaks_count];')
        return self._fetch_state_script_string

    def _refresh_state(self):
        """
        Refresh the snapshot of the underlying video and player state. We do
        this all in one so that the state doesn't change in between queries.

        We also store information that can be derived from the snapshotted
        information, such as lag. This is stored in the last seen state to
        stress that it's based on the snapshot.
        """
        values = self._execute_yt_script(self._fetch_state_script)
        video_keys = ['base_uri', 'current_time', 'duration',
                      'raw_buffered_ranges', 'raw_played_ranges',
                      'total_frames', 'dropped_frames', 'corrupted_frames']
        player_keys = ['player_duration', 'player_current_time',
                       'player_playback_quality', 'player_movie_id',
                       'player_movie_title', 'player_url', 'player_state',
                       'player_ad_state', 'player_breaks_count']
        # Get video state
        self._last_seen_video_state = (
            self._create_video_state_info(**dict(
                zip(video_keys, values[:len(video_keys)]))))
        # Get player state
        self._last_seen_player_state = (
            self._create_player_state_info(**dict(
                zip(player_keys, values[-len(player_keys):]))))

    def mse_enabled(self):
        """
        Check if the video source indicates mse usage for current video.
        Refreshes state.

        :return: True if MSE is being used, False if not.
        """
        self._refresh_state()
        return self._last_seen_video_state.video_src.startswith('blob')

    def playback_started(self):
        """
        Check whether playback has started. Refreshes state.

        :return: True if play back has started, False if not.
        """
        self._refresh_state()
        # usually, ad is playing during initial buffering
        if (self._last_seen_player_state.player_playing or
                self._last_seen_player_state.player_buffering):
            return True
        if (self._last_seen_video_state.current_time > 0 or
                self._last_seen_player_state.player_current_time > 0):
            return True
        return False

    def playback_done(self):
        """
        Check whether playback is done. Refreshes state.

        :return: True if play back has ended, False if not.
        """
        # in case ad plays at end of video
        self._refresh_state()
        if self._last_seen_player_state.player_ad_playing:
            return False
        return (self._last_seen_player_state.player_ended or
                self._last_seen_player_state.player_remaining_time < 1)

    def wait_for_almost_done(self, final_piece=120):
        """
        Allow the given video to play until only `final_piece` seconds remain,
        skipping ads mid-way as much as possible.
        `final_piece` should be short enough to not be interrupted by an ad.

        Depending on the length of the video, check the ad status every 10-30
        seconds, skip an active ad if possible.

        This call refreshes state.

        :param final_piece: The length in seconds of the desired remaining time
        to wait until.
        """
        self._refresh_state()
        rest = 10
        duration = remaining_time = self.expected_duration
        if duration < final_piece:
            # video is short so don't attempt to skip more ads
            return duration
        elif duration > 600:
            # for videos that are longer than 10 minutes
            # wait longer between checks
            rest = duration / 50

        while remaining_time > final_piece:
            if self._player_stalled():
                if self._last_seen_player_state.player_buffering:
                    # fall back on timeout in 'wait' call that comes after this
                    # in test function
                    self.marionette.log('Buffering and no playback progress.')
                    break
                else:
                    message = '\n'.join(['Playback stalled', str(self)])
                    raise VideoException(message)
            if self._last_seen_player_state.player_breaks_count > 0:
                self.process_ad()
            if remaining_time > 1.5 * rest:
                sleep(rest)
            else:
                sleep(rest / 2)
            # TODO during an ad, remaining_time will be based on ad's
            # current_time rather than current_time of target video
            remaining_time = self._last_seen_player_state.player_remaining_time
        return remaining_time

    def __str__(self):
        messages = [super(YouTubePuppeteer, self).__str__()]
        if not self.player:
            messages += ['\t.html5-media-player: None']
            return '\n'.join(messages)
        if not self._last_seen_player_state:
            messages += ['\t.html5-media-player: No last seen state']
            return '\n'.join(messages)
        messages += ['.html5-media-player: {']
        for field in self._last_seen_player_state._fields:
            # For compatibility with different test environments we force ascii
            field_ascii = (
                unicode(getattr(self._last_seen_player_state, field)).encode(
                    'ascii', 'replace'))
            messages += [('\t{}: {}'.format(field, field_ascii))]
        messages += '}'
        return '\n'.join(messages)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os, sys
import glob
import argparse
import traceback
import WebIDL

class TestHarness(object):
    def __init__(self, test, verbose):
        self.test = test
        self.verbose = verbose
        self.printed_intro = False
        self.passed = 0
        self.failures = []

    def start(self):
        if self.verbose:
            self.maybe_print_intro()

    def finish(self):
        if self.verbose or self.printed_intro:
            print "Finished test %s" % self.test

    def maybe_print_intro(self):
        if not self.printed_intro:
            print "Starting test %s" % self.test
            self.printed_intro = True

    def test_pass(self, msg):
        self.passed += 1
        if self.verbose:
            print "TEST-PASS | %s" % msg

    def test_fail(self, msg):
        self.maybe_print_intro()
        self.failures.append(msg)
        print "TEST-UNEXPECTED-FAIL | %s" % msg

    def ok(self, condition, msg):
        if condition:
            self.test_pass(msg)
        else:
            self.test_fail(msg)

    def check(self, a, b, msg):
        if a == b:
            self.test_pass(msg)
        else:
            self.test_fail(msg + " | Got %s expected %s" % (a, b))

def run_tests(tests, verbose):
    testdir = os.path.join(os.path.dirname(__file__), 'tests')
    if not tests:
        tests = glob.iglob(os.path.join(testdir, "*.py"))
    sys.path.append(testdir)

    all_passed = 0
    failed_tests = []

    for test in tests:
        (testpath, ext) = os.path.splitext(os.path.basename(test))
        _test = __import__(testpath, globals(), locals(), ['WebIDLTest'])

        harness = TestHarness(test, verbose)
        harness.start()
        try:
            _test.WebIDLTest.__call__(WebIDL.Parser(), harness)
        except Exception, ex:
            harness.test_fail("Unhandled exception in test %s: %s" %
                              (testpath, ex))
            traceback.print_exc()
        finally:
            harness.finish()
        all_passed += harness.passed
        if harness.failures:
            failed_tests.append((test, harness.failures))

    if verbose or failed_tests:
        print
        print 'Result summary:'
        print 'Successful: %d' % all_passed
        print 'Unexpected: %d' % \
                sum(len(failures) for _, failures in failed_tests)
        for test, failures in failed_tests:
            print '%s:' % test
            for failure in failures:
                print 'TEST-UNEXPECTED-FAIL | %s' % failure
    return 1 if failed_tests else 0

def get_parser():
    usage = """%(prog)s [OPTIONS] [TESTS]
               Where TESTS are relative to the tests directory."""
    parser = argparse.ArgumentParser(usage=usage)
    parser.add_argument('-q', '--quiet', action='store_false', dest='verbose',
                        help="Don't print passing tests.", default=None)
    parser.add_argument('-v', '--verbose', action='store_true', dest='verbose',
                        help="Run tests in verbose mode.")
    parser.add_argument('tests', nargs="*", help="Tests to run")
    return parser

if __name__ == '__main__':
    parser = get_parser()
    args = parser.parse_args()
    if args.verbose is None:
        args.verbose = True

    # Make sure the current directory is in the python path so we can cache the
    # result of the webidlyacc.py generation.
    sys.path.append('.')

    sys.exit(run_tests(args.tests, verbose=args.verbose))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import sys
import string
import argparse
import subprocess
import buildconfig
from mozbuild import shellutil

# Generates a line of WebIDL with the given spelling of the property name
# (whether camelCase, _underscorePrefixed, etc.) and the given array of
# extended attributes.
def generateLine(propName, extendedAttrs):
    return "  [%s] attribute DOMString %s;\n" % (", ".join(extendedAttrs),
                                                 propName)
def generate(output, idlFilename, preprocessorHeader):
    cpp = list(buildconfig.substs['CPP'])
    cpp += shellutil.split(buildconfig.substs['ACDEFINES'])
    cpp.append(preprocessorHeader)
    preprocessed = subprocess.check_output(cpp)

    propList = eval(preprocessed)
    props = ""
    for [name, prop, id, flags, pref, proptype] in propList:
        if "CSS_PROPERTY_INTERNAL" in flags:
            continue
        # Unfortunately, even some of the getters here are fallible
        # (e.g. on nsComputedDOMStyle).
        extendedAttrs = ["Throws", "TreatNullAs=EmptyString"]
        if pref is not "":
            extendedAttrs.append('Pref="%s"' % pref)

        # webkit properties get a capitalized "WebkitFoo" accessor (added here)
        # as well as a camelcase "webkitFoo" accessor (added next).
        if (prop.startswith("Webkit")):
            props += generateLine(prop, extendedAttrs)

        # Generate a line with camelCase spelling of property-name (or capitalized,
        # for Moz-prefixed properties):
        if not prop.startswith("Moz"):
            prop = prop[0].lower() + prop[1:]
        props += generateLine(prop, extendedAttrs)

        # Per spec, what's actually supposed to happen here is that we're supposed
        # to have properties for:
        #
        # 1) Each supported CSS property name, camelCased.
        # 2) Each supported name that contains or starts with dashes,
        #    without any changes to the name.
        # 3) cssFloat
        #
        # Note that "float" will cause a property called "float" to exist due to (1)
        # in that list.
        #
        # In practice, cssFloat is the only case in which "name" doesn't contain
        # "-" but also doesn't match "prop".  So the above generatePropLine() call
        # covered (3) and all of (1) except "float".  If we now output attributes
        # for all the cases where "name" doesn't match "prop", that will cover
        # "float" and (2).
        if prop != name:
            extendedAttrs.append('BinaryName="%s"' % prop)
            # Throw in a '_' before the attribute name, because some of these
            # property names collide with IDL reserved words.
            props += generateLine("_" + name, extendedAttrs)


    idlFile = open(idlFilename, "r")
    idlTemplate = idlFile.read()
    idlFile.close()

    output.write("/* THIS IS AN AUTOGENERATED FILE.  DO NOT EDIT */\n\n" +
                 string.Template(idlTemplate).substitute({"props": props}) + '\n')

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('idlFilename', help='IDL property file template')
    parser.add_argument('preprocessorHeader', help='Header file to pass through the preprocessor')
    args = parser.parse_args()
    generate(sys.stdout, args.idlFilename, args.preprocessorHeader)

if __name__ == '__main__':
    main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from marionette_harness import BrowserMobProxyTestCaseMixin

from external_media_harness.testcase import (
    NetworkBandwidthTestCase, NetworkBandwidthTestsMixin
)


class TestPlaybackLimitingBandwidth(NetworkBandwidthTestCase,
                                    NetworkBandwidthTestsMixin,
                                    BrowserMobProxyTestCaseMixin):

    # Tests are in NetworkBandwidthTestsMixin
    pass



import WebIDL

def WebIDLTest(parser, harness):
    parser.parse("""
        callback TestVariadicCallback = any(any... arguments);
    """)

    results = parser.finish()

    harness.ok(True, "TestVariadicCallback callback parsed without error.")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from marionette_harness import BrowserMobProxyTestCaseMixin

from external_media_harness.testcase import NetworkBandwidthTestCase


class TestUltraLowBandwidth(NetworkBandwidthTestCase,
                                    BrowserMobProxyTestCaseMixin):

    def test_playback_limiting_bandwidth_160(self):
        self.proxy.limits({'downstream_kbps': 160})
        self.run_videos(timeout=120)



#!/usr/bin/env python
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Imports a test suite from a remote repository. Takes one argument, a file in
the format described in README.
Note: removes both source and destination directory before starting. Do not
      use with outstanding changes in either directory.
"""

from __future__ import print_function, unicode_literals

import os
import shutil
import subprocess
import sys

import parseManifest
import writeBuildFiles

def readManifests(iden, dirs):
    def parseManifestFile(iden, path):
        pathstr = "hg-%s/%s/MANIFEST" % (iden, path)
        subdirs, mochitests, reftests, _, supportfiles = parseManifest.parseManifestFile(pathstr)
        return subdirs, mochitests, reftests, supportfiles

    data = []
    for path in dirs:
        subdirs, mochitests, reftests, supportfiles = parseManifestFile(iden, path)
        data.append({
          "path": path,
          "mochitests": mochitests,
          "reftests": reftests,
          "supportfiles": supportfiles,
        })
        data.extend(readManifests(iden, ["%s/%s" % (path, d) for d in subdirs]))
    return data


def getData(confFile):
    """This function parses a file of the form
    (hg or git)|URL of remote repository|identifier for the local directory
    First directory of tests
    ...
    Last directory of tests"""
    vcs = ""
    url = ""
    iden = ""
    directories = []
    try:
        with open(confFile, 'r') as fp:
            first = True
            for line in fp:
                if first:
                    vcs, url, iden = line.strip().split("|")
                    first = False
                else:
                    directories.append(line.strip())
    finally:
        return vcs, url, iden, directories


def makePathInternal(a, b):
    if not b:
        # Empty directory, i.e., the repository root.
        return a
    return "%s/%s" % (a, b)


def makeSourcePath(a, b):
    """Make a path in the source (upstream) directory."""
    return makePathInternal("hg-%s" % a, b)


def makeDestPath(a, b):
    """Make a path in the destination (mozilla-central) directory, shortening as
    appropriate."""
    def shorten(path):
        path = path.replace('dom-tree-accessors', 'dta')
        path = path.replace('document.getElementsByName', 'doc.gEBN')
        path = path.replace('requirements-for-implementations', 'implreq')
        path = path.replace('other-elements-attributes-and-apis', 'oeaaa')
        return path

    return shorten(makePathInternal(a, b))


def extractReftestFiles(reftests):
    """Returns the set of files referenced in the reftests argument"""
    files = set()
    for line in reftests:
        files.update([line[1], line[2]])
    return files


def copy(dest, directories):
    """Copy mochitests and support files from the external HG directory to their
    place in mozilla-central.
    """
    print("Copying tests...")
    for d in directories:
        sourcedir = makeSourcePath(dest, d["path"])
        destdir = makeDestPath(dest, d["path"])
        os.makedirs(destdir)

        reftestfiles = extractReftestFiles(d["reftests"])

        for mochitest in d["mochitests"]:
            shutil.copy("%s/%s" % (sourcedir, mochitest), "%s/test_%s" % (destdir, mochitest))
        for reftest in sorted(reftestfiles):
            shutil.copy("%s/%s" % (sourcedir, reftest), "%s/%s" % (destdir, reftest))
        for support in d["supportfiles"]:
            shutil.copy("%s/%s" % (sourcedir, support), "%s/%s" % (destdir, support))

def printBuildFiles(dest, directories):
    """Create a mochitest.ini that all the contains tests we import.
    """
    print("Creating manifest...")
    all_mochitests = set()
    all_support = set()

    for d in directories:
        path = makeDestPath(dest, d["path"])

        all_mochitests |= set('%s/test_%s' % (d['path'], mochitest)
            for mochitest in d['mochitests'])
        all_support |= set('%s/%s' % (d['path'], p) for p in d['supportfiles'])

        if d["reftests"]:
            with open(path + "/reftest.list", "w") as fh:
                result = writeBuildFiles.substReftestList("importTestsuite.py",
                    d["reftests"])
                fh.write(result)

    manifest_path = dest + '/mochitest.ini'
    with open(manifest_path, 'w') as fh:
        result = writeBuildFiles.substManifest('importTestsuite.py',
            all_mochitests, all_support)
        fh.write(result)
    subprocess.check_call(["hg", "add", manifest_path])

def hgadd(dest, directories):
    """Inform hg of the files in |directories|."""
    print("hg addremoving...")
    for d in directories:
        subprocess.check_call(["hg", "addremove", makeDestPath(dest, d)])

def removeAndCloneRepo(vcs, url, dest):
    """Replaces the repo at dest by a fresh clone from url using vcs"""
    assert vcs in ('hg', 'git')

    print("Removing %s..." % dest)
    subprocess.check_call(["rm", "-rf", dest])

    print("Cloning %s to %s with %s..." % (url, dest, vcs))
    subprocess.check_call([vcs, "clone", url, dest])

def importRepo(confFile):
    try:
        vcs, url, iden, directories = getData(confFile)
        dest = iden
        hgdest = "hg-%s" % iden

        print("Removing %s..." % dest)
        subprocess.check_call(["rm", "-rf", dest])

        removeAndCloneRepo(vcs, url, hgdest)

        data = readManifests(iden, directories)
        print("Going to import %s..." % [d["path"] for d in data])

        copy(dest, data)
        printBuildFiles(dest, data)
        hgadd(dest, directories)
        print("Removing %s again..." % hgdest)
        subprocess.check_call(["rm", "-rf", hgdest])
    except subprocess.CalledProcessError as e:
        print(e.returncode)
    finally:
        print("Done")

if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("Need one argument.")
    else:
        importRepo(sys.argv[1])




import WebIDL

def WebIDLTest(parser, harness):
    try:
        parser.parse("""
            enum TestEnumDuplicateValue {
              "",
              ""
            };
        """)
        harness.ok(False, "Should have thrown!")
    except:
        harness.ok(True, "Enum TestEnumDuplicateValue should throw")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, # You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import, unicode_literals

import os
import sys

from mach.decorators import (
    CommandArgument,
    CommandProvider,
    Command,
)

from mozbuild.base import MachCommandBase
from mozbuild.util import mkdir

def get_test_parser():
    import runtests
    return runtests.get_parser

@CommandProvider
class WebIDLProvider(MachCommandBase):
    @Command('webidl-example', category='misc',
             description='Generate example files for a WebIDL interface.')
    @CommandArgument('interface', nargs='+',
                     help='Interface(s) whose examples to generate.')
    def webidl_example(self, interface):
        from mozwebidlcodegen import BuildSystemWebIDL

        manager = self._spawn(BuildSystemWebIDL).manager
        for i in interface:
            manager.generate_example_files(i)

    @Command('webidl-parser-test', category='testing', parser=get_test_parser,
             description='Run WebIDL tests (Interface Browser parser).')
    def webidl_test(self, **kwargs):
        sys.path.insert(0, os.path.join(self.topsrcdir, 'other-licenses',
                        'ply'))

        # Ensure the topobjdir exists. On a Taskcluster test run there won't be
        # an objdir yet.
        mkdir(self.topobjdir)

        # Make sure we drop our cached grammar bits in the objdir, not
        # wherever we happen to be running from.
        os.chdir(self.topobjdir)

        if kwargs["verbose"] is None:
            kwargs["verbose"] = False

        # Now we're going to create the cached grammar file in the
        # objdir.  But we're going to try loading it as a python
        # module, so we need to make sure the objdir is in our search
        # path.
        sys.path.insert(0, self.topobjdir)

        import runtests
        return runtests.run_tests(kwargs["tests"], verbose=kwargs["verbose"])



def WebIDLTest(parser, harness):
    threw = False
    try:
        parser.parse("""
            interface VariadicConstraints1 {
              void foo(byte... arg1, byte arg2);
            };
        """)
        results = parser.finish()

    except:
        threw = True

    harness.ok(threw,
               "Should have thrown on variadic argument followed by required "
               "argument.")

    parser = parser.reset()
    threw = False
    try:
        parser.parse("""
            interface VariadicConstraints2 {
              void foo(byte... arg1, optional byte arg2);
            };
        """)
        results = parser.finish();
    except:
        threw = True

    harness.ok(threw,
               "Should have thrown on variadic argument followed by optional "
               "argument.")

    parser = parser.reset()
    threw = False
    try:
        parser.parse("""
            interface VariadicConstraints3 {
              void foo(optional byte... arg1);
            };
        """)
        results = parser.finish()

    except:
        threw = True

    harness.ok(threw,
               "Should have thrown on variadic argument explicitly flagged as "
               "optional.")

    parser = parser.reset()
    threw = False
    try:
        parser.parse("""
            interface VariadicConstraints4 {
              void foo(byte... arg1 = 0);
            };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Should have thrown on variadic argument with default value.")



#!/usr/bin/env python
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import unicode_literals

import subprocess

repo = "https://github.com/w3c/testharness.js"
dest = "resources-upstream"
files = [{"f":"testharness.js"},
         {"f":"testharness.css"},
         {"f":"idlharness.js"},
         {"d":"webidl2/lib/webidl2.js", "f":"WebIDLParser.js"}]

subprocess.check_call(["git", "clone", repo, dest])
subprocess.check_call(["git", "submodule", "init"], cwd=dest)
subprocess.check_call(["git", "submodule", "update"], cwd=dest)
for f in files:
    path = f["d"] if "d" in f else f["f"]
    subprocess.check_call(["cp", "%s/%s" % (dest, path), f["f"]])
    subprocess.check_call(["hg", "add", f["f"]])
subprocess.check_call(["rm", "-rf", dest])




def WebIDLTest(parser, harness):
    threw = False
    try:
        parser.parse("""
            [Constructor, NoInterfaceObject]
            interface TestConstructorNoInterfaceObject {
            };
        """)

        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Should have thrown.")

    threw = False
    try:
        parser.parse("""
            [NoInterfaceObject, Constructor]
            interface TestConstructorNoInterfaceObject {
            };
        """)

        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Should have thrown.")

    parser = parser.reset()

    parser.parse("""
        [NoInterfaceObject, NamedConstructor=FooBar]
        interface TestNamedConstructorNoInterfaceObject {
        };
    """)

    # Test HTMLConstructor and NoInterfaceObject
    parser = parser.reset()

    threw = False
    try:
        parser.parse("""
            [NoInterfaceObject, HTMLConstructor]
            interface TestHTMLConstructorNoInterfaceObject {
            };
        """)

        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Should have thrown.")

    parser = parser.reset()

    threw = False
    try:
        parser.parse("""
            [HTMLConstructor, NoInterfaceObject]
            interface TestHTMLConstructorNoInterfaceObject {
            };
        """)

        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Should have thrown.")



def WebIDLTest(parser, harness):
    parser.parse("""
      dictionary Dict2 : Dict1 {
        long child = 5;
        Dict1 aaandAnother;
      };
      dictionary Dict1 {
        long parent;
        double otherParent;
      };
    """)
    results = parser.finish()

    dict1 = results[1];
    dict2 = results[0];

    harness.check(len(dict1.members), 2, "Dict1 has two members")
    harness.check(len(dict2.members), 2, "Dict2 has four members")

    harness.check(dict1.members[0].identifier.name, "otherParent",
                  "'o' comes before 'p'")
    harness.check(dict1.members[1].identifier.name, "parent",
                  "'o' really comes before 'p'")
    harness.check(dict2.members[0].identifier.name, "aaandAnother",
                  "'a' comes before 'c'")
    harness.check(dict2.members[1].identifier.name, "child",
                  "'a' really comes before 'c'")

    # Now reset our parser
    parser = parser.reset()
    threw = False
    try:
        parser.parse("""
          dictionary Dict {
            long prop = 5;
            long prop;
          };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Should not allow name duplication in a dictionary")

    # Now reset our parser again
    parser = parser.reset()
    threw = False
    try:
        parser.parse("""
          dictionary Dict1 : Dict2 {
            long prop = 5;
          };
          dictionary Dict2 : Dict3 {
            long prop2;
          };
          dictionary Dict3 {
            double prop;
          };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Should not allow name duplication in a dictionary and "
               "its ancestor")

    # More reset
    parser = parser.reset()
    threw = False
    try:
        parser.parse("""
          interface Iface {};
          dictionary Dict : Iface {
            long prop;
          };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Should not allow non-dictionary parents for dictionaries")

    # Even more reset
    parser = parser.reset()
    threw = False
    try:
        parser.parse("""
            dictionary A : B {};
            dictionary B : A {};
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Should not allow cycles in dictionary inheritance chains")

    parser = parser.reset()
    threw = False
    try:
        parser.parse("""
            dictionary A {
              [TreatNullAs=EmptyString] DOMString foo;
            };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Should not allow [TreatNullAs] on dictionary members");

    parser = parser.reset()
    threw = False
    try:
        parser.parse("""
            dictionary A {
            };
            interface X {
              void doFoo(A arg);
            };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Trailing dictionary arg must be optional")

    parser = parser.reset()
    threw = False
    try:
        parser.parse("""
            dictionary A {
            };
            interface X {
              void doFoo((A or DOMString) arg);
            };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw,
               "Trailing union arg containing a dictionary must be optional")

    parser = parser.reset()
    threw = False
    try:
        parser.parse("""
            dictionary A {
            };
            interface X {
              void doFoo(A arg1, optional long arg2);
            };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Dictionary arg followed by optional arg must be optional")

    parser = parser.reset()
    threw = False
    try:
        parser.parse("""
            dictionary A {
            };
            interface X {
              void doFoo(A arg1, optional long arg2, long arg3);
            };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(not threw,
               "Dictionary arg followed by non-optional arg doesn't have to be optional")

    parser = parser.reset()
    threw = False
    try:
        parser.parse("""
            dictionary A {
            };
            interface X {
              void doFoo((A or DOMString) arg1, optional long arg2);
            };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw,
               "Union arg containing dictionary followed by optional arg must "
               "be optional")

    parser = parser.reset()
    parser.parse("""
            dictionary A {
            };
            interface X {
              void doFoo(A arg1, long arg2);
            };
        """)
    results = parser.finish()
    harness.ok(True, "Dictionary arg followed by required arg can be required")

    parser = parser.reset()
    threw = False
    try:
        parser.parse("""
            dictionary A {
            };
            interface X {
              void doFoo(optional A? arg1);
            };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Dictionary arg must not be nullable")

    parser = parser.reset()
    threw = False
    try:
        parser.parse("""
            dictionary A {
            };
            interface X {
              void doFoo(optional (A or long)? arg1);
            };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Dictionary arg must not be in a nullable union")

    parser = parser.reset()
    threw = False
    try:
        parser.parse("""
            dictionary A {
            };
            interface X {
              void doFoo(optional (A or long?) arg1);
            };
        """)
        results = parser.finish()
    except:
        threw = True
    harness.ok(threw,
               "Dictionary must not be in a union with a nullable type")

    parser = parser.reset()
    threw = False
    try:
        parser.parse("""
            dictionary A {
            };
            interface X {
              void doFoo(optional (long? or A) arg1);
            };
        """)
        results = parser.finish()
    except:
        threw = True
    harness.ok(threw,
               "A nullable type must not be in a union with a dictionary")

    parser = parser.reset()
    parser.parse("""
        dictionary A {
        };
        interface X {
          A? doFoo();
        };
    """)
    results = parser.finish()
    harness.ok(True, "Dictionary return value can be nullable")

    parser = parser.reset()
    parser.parse("""
        dictionary A {
        };
        interface X {
          void doFoo(optional A arg);
        };
    """)
    results = parser.finish()
    harness.ok(True, "Dictionary arg should actually parse")

    parser = parser.reset()
    parser.parse("""
        dictionary A {
        };
        interface X {
          void doFoo(optional (A or DOMString) arg);
        };
    """)
    results = parser.finish()
    harness.ok(True, "Union arg containing a dictionary should actually parse")

    parser = parser.reset()
    threw = False
    try:
        parser.parse("""
            dictionary Foo {
              Foo foo;
            };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Member type must not be its Dictionary.")

    parser = parser.reset()
    threw = False
    try:
        parser.parse("""
            dictionary Foo3 : Foo {
              short d;
            };

            dictionary Foo2 : Foo3 {
              boolean c;
            };

            dictionary Foo1 : Foo2 {
              long a;
            };

            dictionary Foo {
              Foo1 b;
            };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Member type must not be a Dictionary that "
                      "inherits from its Dictionary.")

    parser = parser.reset()
    threw = False
    try:
        parser.parse("""
            dictionary Foo {
              (Foo or DOMString)[]? b;
            };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Member type must not be a Nullable type "
                      "whose inner type includes its Dictionary.")

    parser = parser.reset()
    threw = False
    try:
        parser.parse("""
            dictionary Foo {
              (DOMString or Foo) b;
            };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Member type must not be a Union type, one of "
                      "whose member types includes its Dictionary.")

    parser = parser.reset()
    threw = False
    try:
        parser.parse("""
            dictionary Foo {
              sequence<sequence<sequence<Foo>>> c;
            };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Member type must not be a Sequence type "
                      "whose element type includes its Dictionary.")

    parser = parser.reset()
    threw = False
    try:
        parser.parse("""
            dictionary Foo {
              (DOMString or Foo)[] d;
            };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Member type must not be an Array type "
                      "whose element type includes its Dictionary.")

    parser = parser.reset()
    threw = False
    try:
        parser.parse("""
            dictionary Foo {
              Foo1 b;
            };

            dictionary Foo3 {
              Foo d;
            };

            dictionary Foo2 : Foo3 {
              short c;
            };

            dictionary Foo1 : Foo2 {
              long a;
            };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Member type must not be a Dictionary, one of whose "
                      "members or inherited members has a type that includes "
                      "its Dictionary.")

    parser = parser.reset();
    threw = False
    try:
        parser.parse("""
            dictionary Foo {
            };

            dictionary Bar {
              Foo? d;
            };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Member type must not be a nullable dictionary")

    parser = parser.reset();
    parser.parse("""
        dictionary Foo {
          unrestricted float  urFloat = 0;
          unrestricted float  urFloat2 = 1.1;
          unrestricted float  urFloat3 = -1.1;
          unrestricted float? urFloat4 = null;
          unrestricted float  infUrFloat = Infinity;
          unrestricted float  negativeInfUrFloat = -Infinity;
          unrestricted float  nanUrFloat = NaN;

          unrestricted double  urDouble = 0;
          unrestricted double  urDouble2 = 1.1;
          unrestricted double  urDouble3 = -1.1;
          unrestricted double? urDouble4 = null;
          unrestricted double  infUrDouble = Infinity;
          unrestricted double  negativeInfUrDouble = -Infinity;
          unrestricted double  nanUrDouble = NaN;
        };
    """)
    results = parser.finish()
    harness.ok(True, "Parsing default values for unrestricted types succeeded.")

    parser = parser.reset();
    threw = False
    try:
        parser.parse("""
            dictionary Foo {
              double f = Infinity;
            };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Only unrestricted values can be initialized to Infinity")

    parser = parser.reset();
    threw = False
    try:
        parser.parse("""
            dictionary Foo {
              double f = -Infinity;
            };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Only unrestricted values can be initialized to -Infinity")

    parser = parser.reset();
    threw = False
    try:
        parser.parse("""
            dictionary Foo {
              double f = NaN;
            };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Only unrestricted values can be initialized to NaN")

    parser = parser.reset();
    threw = False
    try:
        parser.parse("""
            dictionary Foo {
              float f = Infinity;
            };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Only unrestricted values can be initialized to Infinity")


    parser = parser.reset();
    threw = False
    try:
        parser.parse("""
            dictionary Foo {
              float f = -Infinity;
            };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Only unrestricted values can be initialized to -Infinity")

    parser = parser.reset();
    threw = False
    try:
        parser.parse("""
            dictionary Foo {
              float f = NaN;
            };
        """)
        results = parser.finish()
    except:
        threw = True

    harness.ok(threw, "Only unrestricted values can be initialized to NaN")
